"""
Provides the Dataset class
"""

from __future__ import annotations

from typing import Optional
import numpy as np
import scipy.sparse as sparse

from .typing import Modality, Axis, DataTensor
from .extras.prior import Prior
from .numbafied import extract_d_values

class Dataset:
    def __init__(
        self,
        *,
        dataset: dict[Modality, DataTensor],
        structure: dict[Modality, tuple[Axis]],
        batch_axes: Optional[set[Axis]] = None
    ):
        # Type declarations of attributes
        self.dataset: dict[Modality, DataTensor]
        self.structure: dict[Modality, tuple[Axis]]

        self.batch_axes: set[Axis]
        self.presences: dict[Axis, dict[Modality, Optional[int]]]
        self.presences_batchless: dict[Axis, dict[Modality, Optional[int]]]

        self.full_sizes: dict[Modality, np.ndarray]
        self.left_sizes: dict[Modality, np.ndarray]
        self.right_sizes: dict[Modality, np.ndarray]
        self.Ks: dict[str, int]

        self.axis_sizes: dict[Axis, int]

        # Type declarations of cached attributes
        # calculated via sklearn api calls
        self.prior: dict[Axis, Prior]
        self.mean: dict[Axis, float]
        self.gram_matrices: dict[Axis, np.ndarray]
        self.es: dict[Axis, np.ndarray]
        self.evecs: dict[Axis, np.ndarray]
        self.evals: dict[Axis, np.ndarray]
        self.precision_matrices: dict[Axis, np.ndarray]

        # Input handling for default arguments
        if batch_axes is None:
            self.batch_axes = set({""})
        else:
            self.batch_axes = batch_axes

        # Set attributes
        self.dataset = dataset
        self.structure = structure
        self.gram_matrices = {}
        self.prior = {}
        self.es = {}
        self.evecs = {}
        self.evals = {}
        self.precision_matrices = {}

        # Calculate helper attributes
        self.recompute_helpers()

        # Make cached attributes for covariance subdivision
        self.cache()

    def memory_usage(self) -> tuple[int, int, int]:
        """
        Returns the memory usage of the dataset
        """
        dataset_bytes: int = 0
        aux_bytes: int = 0
        precision_bytes: int = 0

        # Keep track of dataset size
        if self.old_dataset is self.dataset:
            for tensor in self.dataset.values():
                dataset_bytes += array_bytes(tensor)
        else:
            # If parts of the dataset has been cached,
            # report the cached dataset as the uncached bits
            # are views, i.e. they do not take up any memory
            for tensor in self.old_dataset.values():
                dataset_bytes += array_bytes(tensor)

        # Add all intermediat results
        for gram_matrix in self.gram_matrices.values():
            aux_bytes += array_bytes(gram_matrix)
        for evals in self.evals.values():
            aux_bytes += array_bytes(evals)
        for evecs in self.evecs.values():
            aux_bytes += array_bytes(evecs)

        # Add output
        for precision_matrix in self.precision_matrices.values():
            precision_bytes += array_bytes(precision_matrix)

        # Potentially add cached gram matrices
        if self.gram_matrices is not self.old_gram_matrices:
            for gram_matrix in self.old_gram_matrices.values():
                aux_bytes += array_bytes(gram_matrix)

        return dataset_bytes, aux_bytes, precision_bytes
    
    def print_memory_usage(self) -> None:
        """
        Prints the memory usage of the dataset
        """
        dataset_bytes, aux_bytes, precision_bytes = self.memory_usage()
        print(f"Dataset: {dataset_bytes / 1e9:.2f} GB")
        print(f"Auxiliary: {aux_bytes / 1e9:.2f} GB")
        print(f"Output: {precision_bytes / 1e9:.2f} GB")

    def cache(self) -> None:
        self.old_dataset = self.dataset
        self.old_structure = self.structure
        self.old_all_axes = self.all_axes
        self.old_axis_sizes = self.axis_sizes
        self.old_gram_matrices = self.gram_matrices

    def uncache(self) -> None:
        self.dataset = self.old_dataset
        self.structure = self.old_structure
        self.gram_matrices = self.old_gram_matrices
        self.recompute_helpers()


    def recompute_helpers(self) -> None:
        self._calculate_axes()
        self._calculate_presences()
        self._calculate_dimensions()
        self._calculate_axis_sizes()

    def _calculate_axes(self) -> None:
        """
        Gets a set of every axis in the dataset, i.e.
        [("A", "B"), ("A", "B")]
        => {"A", "B", "C"}
        """
        self.all_axes: set[str] = {
            axis
            for axes in self.structure.values()
            for axis in axes
        } - self.batch_axes

    def _calculate_presences(self) -> None:
        """
        Returns a dictionary keyed by axis names,
        indicating which datasets the axis is present in

        We assume each axis appears at most once per dataset
        """

        self.presences: dict[Axis, dict[Modality, Optional[int]]] = {}
        self.presences_batchless: dict[Axis, dict[Modality, Optional[int]]] = {}

        for axis in self.all_axes:
            self.presences[axis] = {}
            self.presences_batchless[axis] = {}
            for modality, axes in self.structure.items():
                to_add: list[int] = [
                    idx
                    for idx, ax
                    in enumerate(axes)
                    if ax == axis
                ]

                if len(to_add) == 0:
                    self.presences[axis][modality] = None
                    self.presences_batchless[axis][modality] = None
                elif len(to_add) == 1:
                    presence: int = to_add[0]
                    batches_before_presence: int = len(
                        [
                            ax
                            for ax in axes[:presence]
                            if ax in self.batch_axes
                        ]
                    )
                    self.presences[axis][modality] = presence
                    self.presences_batchless[axis][modality] = presence - batches_before_presence
                else:
                    raise Exception("GmGM does not allow repeated axes in a dataset.")
                
    def _calculate_dimensions(self) -> None:
        """
        Get d_\forall, d_{<\ell}, and d_{>\ell}, and K for each modality,
        as defined in the paper.
        """
        self.full_sizes: dict[Modality, np.ndarray] = {}
        self.left_sizes: dict[Modality, np.ndarray] = {}
        self.right_sizes: dict[Modality, np.ndarray] = {}

        for modality, tensor in self.dataset.items():
            full_size, left_size, right_size = extract_d_values(
                tensor.shape
            )
            self.full_sizes[modality] = full_size
            self.left_sizes[modality] = left_size
            self.right_sizes[modality] = right_size

        # Get the number of dimensions for each tensor
        self.Ks: dict[str, int] = {
            modality: len(
                set(self.structure[modality]) - self.batch_axes
            )
            for modality in self.dataset.keys()
        }

    def _calculate_axis_sizes(self) -> None:
        """
        Calculates the size of each axis
        """
        self.axis_sizes: dict[Axis, int] = {}
        for axis in self.all_axes:
            self.axis_sizes[axis] = [
                tensor.shape[idx]
                for modality, tensor in self.dataset.items()
                if (idx := self.presences[axis][modality]) is not None
            ][0]

    def __repr__(self) -> str:

        # Print structure of each dataset
        dataset_str: str = "\n".join(
            f"\t{modality}: {axes}"
            for modality, axes in self.structure.items()
        )

        # Print calculation status info
        prior_str: dict[Axis, str] = {
            axis: "Prior: " + (self.prior[axis].name if axis in self.prior else "None")
            for axis in self.all_axes
        }
        gram_str: dict[Axis, str] = {
            axis: "Gram: " + ("Calculated" if axis in self.gram_matrices else "Not calculated")
            for axis in self.all_axes
        }
        eig_str: dict[Axis, str] = {
            axis: "Eig: " + ("Calculated" if axis in self.evecs else "Not calculated")
            for axis in self.all_axes
        }

        # Print the size of each axis
        size_str: str = "\n".join(
            f"\t{axis}: {size}\n\t\t"
            + "\n\t\t".join([x[axis] for x in [prior_str, gram_str, eig_str]])
            for axis, size in self.axis_sizes.items()
        )

        return f"Dataset(\n{dataset_str}\n)\nAxes(\n{size_str}\n)"
    
def array_bytes(
    arr: np.ndarray | sparse.sparray
) -> float:
    """
    Returns the number of bytes required to store
    a numpy array or sparse matrix
    """
    if isinstance(arr, sparse.spmatrix):
        raise ValueError(
            "Please use sparse arrays instead of the to-be-deprecated"
            + " sparse matrix API."
        )
    
    # Input handling for various sparse array types
    if isinstance(arr, sparse.csr_array) | isinstance(arr, sparse.csc_array):
        return (
            arr.data.nbytes
            + arr.indices.nbytes
            + arr.indptr.nbytes
        )
    elif isinstance(arr, sparse.coo_array):
        return (
            arr.data.nbytes
            + arr.row.nbytes
            + arr.col.nbytes
        )
    else:
        return arr.nbytes


from typing import *
import numpy as np
import matplotlib.pyplot as plt
from .GmGM import *
from .extras.regularizers import *
from .typing import *
from .generate_data import *

class RunningMeasurer:
    def __init__(self):
        self.mean = 0
        self.var = 0
        self.n = 0

    def __call__(self, input):
        self.mean = (self.mean * self.n + input) / (self.n + 1)
        self.var = (self.var * self.n + (input - self.mean) ** 2) / (self.n + 1)
        self.n += 1
    
    def std(self):
        return np.sqrt(self.var)
    
AlgorithmName: TypeAlias = str
RegularizationParameter: TypeAlias = float
PrecisionMatrix: TypeAlias = np.ndarray
AxisName: TypeAlias = str
MetricName: TypeAlias = str

Algorithm: TypeAlias = Callable[
    [Dataset, RegularizationParameter],
    PrecisionMatrix
]

def measure_prec_recall(
    generator: DatasetGenerator,
    algorithms: dict[AlgorithmName, Algorithm],
    Λs: dict[Algorithm, list[float]],
    num_attempts: int,
    num_samples: int = 100,
    *,
    verbose: int = 0,
    give_prior: bool = False
) -> dict[
    AlgorithmName,
    dict[
        AxisName,
        dict[
            MetricName,
            list[float]
        ]
    ]
]:
    """
    Using `generator`, test the performance of `algorithm` on the dataset
    as the regularization parameter Λ varies

    For each Λ, we run `num_attempts` attempts and average the results
    """

    random_alg = list(algorithms.keys())[0]
    output = [None] * len(Λs[random_alg])

    num_λs = len(Λs[random_alg])

    # Create measurers
    measurers = []
    for idx in range(num_λs):
        measurers.append({
            algorithm_name: {
                axis_name: {
                    metric_name: RunningMeasurer()
                    for metric_name in ["precision", "recall"]
                }
                for axis_name in generator.axes
            }
            for algorithm_name in algorithms.keys()
        })

    for i in range(num_attempts):
        if verbose >= 1:
            print(f"Attempt {i+1}/{num_attempts}")

        # Generate a new ground truth
        generator.reroll_Ψs()
        Ψs_true = generator.Ψs

        # Use this new ground truth to generate
        # an input dataset
        dataset = generator.generate(num_samples)

        for idx in range(num_λs):
            if verbose >= 2:
                print(f"λ #{idx}")

            # For each algorithm collect metrics for that
            # algorithm on this dataset
            for algorithm_name, algorithm in algorithms.items():
                if verbose >= 3:
                    print(f"Algorithm: {algorithm_name}")

                # Run algorithm
                if not give_prior:
                    Ψs_pred = algorithm(dataset, generator.structure, Λs[algorithm_name][idx])
                else:
                    Ψs_pred = algorithm(dataset, generator.structure, Λs[algorithm_name][idx], Ψs_true)

                # Get metrics
                Ψs_pred = binarize_matrices(Ψs_pred, eps=1e-3, mode="<Tolerance")
                cm = {
                    axis: generate_confusion_matrices(Ψs_pred[axis], Ψs_true[axis])
                    for axis in generator.axes
                    if axis in Ψs_pred
                }

                precisions = {
                    axis: precision(cm[axis])
                    if axis in cm
                    else 0
                    for axis in generator.axes
                }
                recalls = {
                    axis: recall(cm[axis])
                    if axis in cm
                    else 0
                    for axis in generator.axes
                }

                # Keep count in a Running Measurer
                for axis in generator.axes:
                    measurers[idx][algorithm_name][axis]["precision"](precisions[axis])
                    measurers[idx][algorithm_name][axis]["recall"](recalls[axis])

    for idx in range(num_λs):
        # Get results from the running measurer into a nice dictionary format
        output[idx] = {
            algorithm_name: {
                axis_name: {
                    metric_name: measurers[idx][algorithm_name][axis_name][metric_name].mean
                    for metric_name in ["precision", "recall"]
                }
                for axis_name in generator.axes
            }
            for algorithm_name in algorithms.keys()
        }

        # Add the standated deviations into precision_std and recall_std keys
        for algorithm_name in algorithms.keys():
            for axis_name in generator.axes:
                for metric_name in ["precision", "recall"]:
                    output[idx][algorithm_name][axis_name][metric_name + "_std"] = \
                        measurers[idx][algorithm_name][axis_name][metric_name].std()

    return output

def plot_prec_recall(
    results: dict,
    axis: str,
    generator: DatasetGenerator
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plots results of `measure_prec_recall` for `axis`
    """

    # Only keep results for the correct axis
    results = [
        {
            algorithm_name: {
                metric_name: metric_results
                for metric_name, metric_results in algorithm_results[axis].items()
            }
            for algorithm_name, algorithm_results in result.items()
        } for result in results
    ]

    # Now plot the results
    fig, ax = plt.subplots()

    for algorithm_name, _ in results[0].items():
        # Plot each algorithms PR curve
        precisions = [result[algorithm_name]["precision"] for result in results]
        recalls = [result[algorithm_name]["recall"] for result in results]

        ax.plot(recalls, precisions, label=algorithm_name)

        # Add error bounds
        precisions_std = [result[algorithm_name]["precision_std"] for result in results]
        recalls_std = [result[algorithm_name]["recall_std"] for result in results]

        ax.fill_between(
            recalls,
            np.array(precisions) - np.array(precisions_std),
            np.array(precisions) + np.array(precisions_std),
            alpha=0.2
        )

        ax.fill_between(
            np.array(recalls) - np.array(recalls_std),
            precisions,
            precisions,
            alpha=0.2
        )

    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    ax.set_xlabel("Recall")
    ax.set_ylabel("Precision")
    ax.legend()

    if generator.name is not None:
        ax.set_title(f"Precision-Recall for {generator.name} on {axis}")

    return fig, ax

def generate_confusion_matrices(
    pred: np.ndarray,
    truth: np.ndarray,
    eps: float = 0,
    mode: Literal[
        '<Tolerance',
        'Nonzero',
        'Negative',
        'Mixed',
    ] = '<Tolerance'
) -> np.ndarray:
    """
    `mode`:
        '<Tolerance': x->1 if |x| < eps, else x->0
        'Nonzero': Same as '<Tolerance'
        'Negative': x->1 if x < 0 else x -> 0
        'Mixed': truth is '<Tolerance', pred is 'Negative'
    """
    mode_pred = mode if mode != 'Mixed' else 'Negative'
    mode_truth = mode if mode != 'Mixed' else 'Nonzero'
    pred = binarize_matrix(pred, eps=eps, mode=mode_pred)
    truth = binarize_matrix(truth, eps=0, mode=mode_truth)
    
    # Identity matrices to remove diagonals
    In = np.eye(pred.shape[0])
    
    np.fill_diagonal(pred, 1)
    np.fill_diagonal(truth, 1)
    
    TP: "True positives"
    TP = (pred * truth - In).sum()
    
    FP: "False positives"
    FP = (pred * (1 - truth)).sum()
    
    TN: "True negatives"
    TN = ((1 - pred) * (1 - truth)).sum()
    
    FN: "False negatives"
    FN = ((1 - pred) *  truth).sum()
    
    return np.array([
        [TP, FP],
        [FN, TN]
    ])

def precision(
    cm: np.ndarray
):
    """
    [[TP, FP], [FN, TN]]
    """
    denom = (cm[0, 0] + cm[0, 1])
    if denom == 0:
        return 1
    return cm[0, 0] / denom

def recall(
    cm: np.ndarray
):
    """
    [[TP, FP], [FN, TN]]
    """
    denom = (cm[0, 0] + cm[1, 0])
    if denom == 0:
        return 1
    return cm[0, 0] / denom

def binarize_matrix(
    M: np.ndarray,
    eps: float = 0,
    mode: Literal["Negative", "<Tolerance"] = '<Tolerance'
):
    """
    Returns M but with only ones and zeros
    """
    out = np.empty(M.shape)
    if mode == '<Tolerance' or mode == 'Nonzero':
        out[np.abs(M) <= eps] = 0
        out[np.abs(M) > eps] = 1
    elif mode == 'Negative':
        out[M < 0] = 1
        out[M >= 0] = 0
        
        # In negative mode we need to add diagonals back
        out += np.eye(out.shape[0])
    else:
        raise Exception(f'Invalid mode {mode}')
    return out

def binarize_matrices(
    Ms: dict[Axis],
    eps: float = 0,
    mode: Literal["Negative", "<Tolerance"] = '<Tolerance'
):
    return {
        axis: binarize_matrix(M, eps=eps, mode=mode)
        for axis, M in Ms.items()
    }


"""
This file contains the methods that are meant to be used
before eigendecomposition/singular value decomposition

Specifically, this includes centering, grammifying, 
and nonparanormal-skeptic-related methods.
"""

from __future__ import annotations
from typing import Optional

from ..dataset import Dataset
from ..typing import Axis
import scipy.stats as stats
import numpy as np

def center(
    X: Dataset
) -> Dataset:
    """
    Modifies X in place
    """

    for modality in X.dataset.keys():
        X.dataset[modality] -= X.dataset[modality].mean()

    return X

def grammify(
    X: Dataset,
    batch_size: Optional[int] = None,
    use_nonparanormal_skeptic: bool = False
) -> Dataset:
    """
    Modifies X in place
    """
    # Clear old gram matrices
    X.gram_matrices = {}

    # Calculate effective gram matrices
    num_samples: dict[Axis, int] = {}
    
    for modality, tensor in X.dataset.items():
        for idx, axis in enumerate(X.structure[modality]):
            if axis in X.batch_axes:
                continue

            matricized: np.ndarray = np.reshape(
                np.moveaxis(tensor, idx, 0),
                (tensor.shape[idx], -1),
                #order='F'
            )

            if axis in X.gram_matrices:
                X.gram_matrices[axis] += _grammify_core(
                    matricized,
                    batch_size,
                    use_nonparanormal_skeptic
                )
                num_samples[axis] += X.full_sizes[modality] // X.axis_sizes[axis]
            else:
                X.gram_matrices[axis] = _grammify_core(
                    matricized,
                    batch_size,
                    use_nonparanormal_skeptic
                )
                num_samples[axis] = X.full_sizes[modality] // X.axis_sizes[axis]

    # Divide gram matrices by effective number of modalities of samples seen
    for axis in X.gram_matrices.keys():
        X.gram_matrices[axis] /= num_samples[axis]

    return X

def _grammify_core(
    matricized: np.ndarray,
    batch_size: Optional[int] = None,
    use_nonparanormal_skeptic: bool = False
) -> np.ndarray:
    """
    TODO: Check if this still works
    """
    
    output: np.ndarray
    if batch_size is None:
        batch_size = matricized.shape[0]

    if use_nonparanormal_skeptic:
        for idx in range(0, matricized.shape[0] + batch_size, batch_size):
            increase = min(batch_size, matricized.shape[0] - idx)
            if idx <= 0:
                break

            matricized[idx:idx+increase] = stats.rankdata(matricized[idx:idx+increase], axis=0)
            matricized[idx:idx+increase] -= matricized[idx:idx+increase].mean(axis=0, keepdims=True)

    output = matricized @ matricized.T
    if use_nonparanormal_skeptic:
        output = np.sin(np.pi/6 * output)
    
    return output



from __future__ import annotations

import numpy as np

class Regularizer:
    """
    A subclassable regularizer class.
    
    Has a method to get the gradient the regularizer
        on a set of eigenvalues and eigenvectors.
        and a method to get the loss of the regularizer.

    """

    def __init__(
        self,
        rhos: dict[str, np.ndarray]
    ):
        """
        Initializes the regularizer with a set of penalties `rhos`.
        """
        self.rhos = rhos

    def loss(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> float:
        """
        Returns the loss of the regularizer
            on a set of eigenvalues and eigenvectors.
        """
        pass

    def grad(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> dict[str, np.ndarray]:
        """
        Returns the gradient of the regularizer
            on a set of eigenvalues and eigenvectors.
        """
        pass

class L1(Regularizer):
    """
    Restricted L1 Regularizer.
    """

    def __init__(
        self,
        rhos: dict[str, np.ndarray]
    ):
        super().__init__(rhos)

        # Keep track of reconstructed matrices
        #  to avoid recomputing them.

        self.reconstructions: np.ndarray = {}

    def loss(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> float:
        to_return = 0
        for axis, recon in self.reconstructions.items():
            to_return += self.rhos[axis] * np.abs(recon).sum()
        return to_return

    def grad(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> dict[str, np.ndarray]:
        to_return: dict[str, np.ndarray] = {}
        for axis in evals.keys():
            if self.rhos[axis] == 0:
                to_return[axis] = np.zeros_like(evals[axis])
                continue
            self.reconstructions[axis] = (
                (evecs[axis] * evals[axis]) @ evecs[axis].T
            )
            np.fill_diagonal(self.reconstructions[axis], 0)
            core: np.ndarray = np.sign(
                self.reconstructions[axis]
            )
            to_return[axis] = self.rhos[axis] * np.einsum(
                "ij, ia, ja -> a",
                core,
                evecs[axis],
                evecs[axis],
                optimize=True
            )
        return to_return
    
class L1Approx(Regularizer):
    """
    L1 Regularizer using an approximate reconstruction
    """

    def __init__(
        self,
        rhos: dict[str, np.ndarray],
        num_eigs: int = 1
    ):
        super().__init__(rhos)

        self.num_eigs = num_eigs

        # Keep track of reconstructed matrices
        #  to avoid recomputing them.
        self.reconstructions: np.ndarray = {}

    def loss(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> float:
        to_return = 0
        for axis, recon in self.reconstructions.items():
            to_return += self.rhos[axis] * np.abs(recon).sum()
        return to_return

    def grad(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> dict[str, np.ndarray]:
        to_return: dict[str, np.ndarray] = {}
        for axis in evals.keys():
            if self.rhos[axis] == 0:
                to_return[axis] = np.zeros_like(evals[axis])
                continue
            # Grab the top `num_eigs` eigenvalues and eigenvectors
            #  and reconstruct the matrix from them.
            top_idx = np.argpartition(evals[axis], -self.num_eigs)[-self.num_eigs:]

            top_evecs = evecs[axis][:, top_idx]
            top_evals = evals[axis][top_idx]

            self.reconstructions[axis] = (
                (top_evecs * top_evals) @ top_evecs.T
            )

            np.fill_diagonal(self.reconstructions[axis], 0)
            core: np.ndarray = np.sign(
                self.reconstructions[axis]
            )
            to_return[axis] = np.zeros_like(evals[axis])
            to_return[axis][top_idx] = self.rhos[axis] * np.einsum(
                "ij, ia, ja -> a",
                core,
                top_evecs,
                top_evecs,
                optimize=True
            )

        return to_return
    
class L2(Regularizer):
    """
    L2 Regularizer

    This is same as the Frobenius norm.
    Hence, we can just take the magnitude of
        the eigenvalues!
    """

    def __init__(
        self,
        rhos: dict[str, np.ndarray]
    ):
        super().__init__(rhos)

        # Keep track of magnitudes to avoid recomputations
        self.magnitudes: dict[str, float] = {}
    

    def loss(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> float:
        to_return = 0
        for axis in evals.keys():
            if self.rhos[axis] == 0:
                continue
            to_return += self.rhos[axis] * self.magnitudes[axis]
        return to_return
    
    def grad(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> dict[str, np.ndarray]:
        to_return: dict[str, np.ndarray] = {}
        for axis in evals.keys():
            if self.rhos[axis] == 0:
                to_return[axis] = np.zeros_like(evals[axis])
                continue
            self.magnitudes[axis] = np.linalg.norm(evals[axis])
            to_return[axis] = self.rhos[axis] / self.magnitudes[axis] * evals[axis]
        return to_return
    
class Random(Regularizer):
    """
    Not actually a regularizer, just adds randomness to the gradient.

    Does poorly.
    """

    def __init__(
        self,
        rhos: dict[str, np.ndarray]
    ):
        super().__init__(rhos)

    def loss(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> float:
        return 0

    def grad(
        self,
        evals: dict[str, np.ndarray],
        evecs: dict[str, np.ndarray],
    ) -> dict[str, np.ndarray]:
        to_return: dict[str, np.ndarray] = {}
        for axis in evals.keys():
            to_return[axis] = self.rhos[axis] * np.random.random(evals[axis].shape)
        return to_return


import numpy as np
import scipy.sparse as sparse

try:
    # TypeAlias only added in Python 3.10
    from typing import TypeAlias
except ImportError:
    TypeAlias = "TypeAlias"

Axis: TypeAlias = str
Modality: TypeAlias = str
DataTensor: TypeAlias = np.ndarray

"""
In python 3.12, this becomes:

type Axis = str
type Modality = str
type DataTensor = np.ndarray
"""


"""
This file contains the methods corresponding to the covariance
subdivision trick (Theorem 3 in the ArXiv paper for GmGM)
"""

from __future__ import annotations
from typing import Optional

from ..typing import DataTensor
from ..dataset import Dataset, Modality, Axis
import scipy.sparse as sparse
import numpy as np

# TODO: Replace this dependency with sparse.csgraph
import igraph as ig

# TODO: Create context managers for these

def subdivide_covariance(
    X: Dataset,
    threshold: Optional[dict[Axis, float] | float],
    *,
    axis_separator: str = "||",
    min_component_size: int = 2,
    verbose: bool = False,
) -> Dataset:
    """
    Uses Theorem 3 in the GmGM ArXiv paper to reduce the problem size
    """
    X.cache()

    # Do covariance partitioning
    components_map: dict[Axis, ig.VertexClustering] = {}
    to_skip: dict[Axis, set[int]] = {axis: set({}) for axis in X.all_axes}
    new_dataset: dict[Modality, DataTensor] = dict({})
    new_grams: dict[Axis, np.ndarray] = dict({})
    new_structure: dict[Modality, tuple[Axis, ...]] = dict({})

    if threshold is None:
        threshed_gram: dict[Axis, np.ndarray] = X.gram_matrices
    elif isinstance(threshold, float):
        threshed_gram: dict[Axis, np.ndarray] = {
            axis: Threshold.shrink_axis(X.gram_matrices[axis], threshold, safe=True)
            for axis in X.gram_matrices.keys()
        }
    else:
        threshed_gram: dict[Axis, np.ndarray] = {
            axis: Threshold.shrink_axis(X.gram_matrices[axis], threshold[axis], safe=True)
            for axis in X.gram_matrices.keys()
        }
    if self.verbose:
        print("Finished thresholding")

    sep: str = axis_separator

    for axis in X.all_axes:
        if verbose:
            print(f"{axis=}")
        if axis in X.batch_axes:
            continue
        for modality, data in X.dataset.items():
            dimension = X.presences[axis][modality]
            gram_matrix: np.ndarray = threshed_gram[axis]
            thresh: np.ndarray = (gram_matrix + gram_matrix.T) != 0
            if isinstance(thresh, sparse.sparray):
                # igraph does not yet support scipy's sparse array API
                # TODO: Migrate to just using pure scipy as there is a
                # scipy.sparse.csgraph.connected_components function
                # TODO: Allow user defined backends?
                #raise NotImplementedError("igraph does not yet support scipy's sparse array API")
                thresh = sparse.csr_matrix(thresh)
            graph: ig.Graph = ig.Graph.Adjacency(thresh).as_undirected()
            if verbose:
                print("Created Graph")

            # Get connected components
            components = graph.components()
            components_map[axis] = components

            if verbose:
                print(f"Got {len(components)} connected components")
                largest_component_size: int = max(len(component) for component in components)
                print(f"The largest component has size {largest_component_size}")
            for idx, component in enumerate(components):
                if len(component) < min_component_size:
                    to_skip[axis].add(idx)
                    continue

                if isinstance(gram_matrix, sparse.sparray | sparse.spmatrix):
                    new_grams[f'{axis}{sep}{idx}'] = gram_matrix[np.ix_(component, component)]
                else:
                    # coo does not support indexing!
                    new_grams[f'{axis}{sep}{idx}'] = gram_matrix.tocsr()[np.ix_(component, component)].tocoo()

                if isinstance(data, sparse.sparray | sparse.spmatrix):
                    # Recall that sparse arrays don't support batches, so it
                    # is either first or second axis
                    batchless_pres = X.presences_batchless[axis][modality]
                    if batchless_pres == 0:
                        new_dataset[f'{modality}{sep}{idx}'] = data[component, :]
                    elif batchless_pres == 1:
                        new_dataset[f'{modality}{sep}{idx}'] = data[:, component]
                    else:
                        raise Exception("Something interesting went wrong...")
                else:
                    new_dataset[f'{modality}{sep}{idx}'] = data.take(component, axis=dimension)
                new_structure[f'{modality}{sep}{idx}'] = list(X.structure[modality])
                new_structure[f'{modality}{sep}{idx}'][dimension] = f'{axis}||{idx}'
                new_structure[f'{modality}{sep}{idx}'] = tuple(new_structure[f'{modality}||{idx}'])

        X.dataset = new_dataset
        X.structure = new_structure
        X.recompute_helpers()

        new_dataset = dict({})
        new_structure = dict({})

    X.gram_matrices = new_grams

    X.components_map = components_map
    X.to_skip = to_skip
    if verbose:
        print("Final keys:")
        print(f"\t{set(X.gram_matrices.keys())}")
    return X

def recombine_precisions(
    X: Dataset,
    axis_separator: str = "||",
    verbose: bool = False
) -> Dataset:
    """
    Modifies X in place

    TODO: Make this more efficient
    TODO: Check if this still works
    """
    batch_size: int = 1
    new_precision_matrices: dict[Axis, np.ndarray] = {}

    for axis, sz in X.dataset.old_axis_sizes.items():
        new_precision_matrices[axis] = \
            sparse.dok_array((sz, sz), dtype=np.float32)

    for axis, Psi in X.precision_matrices.items():
        if verbose:
            print(axis)
        original_axis: Axis = axis.split(axis_separator)[0]
        group_id: int = int(axis.split(axis_separator)[1])

        if group_id in X.to_skip[original_axis]:
            continue
        original_locations = np.where(
            np.array(X.components_map[original_axis].membership) == group_id
        )[0]
        for idx in range(0, len(original_locations), batch_size):
            if verbose:
                print(idx)
            # Do it row-wise because I think scipy creates a dense matrix
            # when you do bulk indexing
            # (Trying to assign it makes me run out of memory if I do it)
            next_ = min(idx + batch_size, len(original_locations))
            rows = original_locations[idx:next_]
            new_precision_matrices[original_axis][
                np.ix_(rows, original_locations)
            ] = Psi[idx:next_, :]

    X.precision_matrices = new_precision_matrices
    return X